@@ -20,11 +20,13 @@
 #include <sys/stat.h>
 #include <sys/time.h>
 #include <time.h>
+#include <limits.h>
 
 #include "nls.h"
 #include "blkdev.h"
 #include "common.h"
 #include "fdisk.h"
+#include "wholedisk.h"
 
 #include "fdisksunlabel.h"
 #include "fdisksgilabel.h"
@@ -211,11 +213,12 @@ unsigned long long sector_offset = 1, extended_offset = 0, sectors;
 unsigned int	heads,
 	cylinders,
 	sector_size = DEFAULT_SECTOR_SIZE,
+	sector_factor = 1,
 	user_set_sector_size = 0,
 	units_per_sector = 1,
 	display_in_cyl_units = 1;
 
-unsigned long long total_number_of_sectors;
+unsigned long long total_number_of_sectors;	/* (!) 512-byte sectors */
 
 #define dos_label (!sun_label && !sgi_label && !aix_label && !mac_label && !osf_label)
 int	sun_label = 0;			/* looking at sun disklabel */
@@ -672,6 +675,21 @@ warn_cylinders(void) {
 "2) booting and partitioning software from other OSs\n"
 "   (e.g., DOS FDISK, OS/2 FDISK)\n"),
 			cylinders);
+
+	if (total_number_of_sectors > UINT_MAX) {
+		int giga = (total_number_of_sectors << 9) / 1000000000;
+		int hectogiga = (giga + 50) / 100;
+
+		fprintf(stderr, _("\n"
+"WARNING: The size of this disk is %d.%d TB (%llu bytes).\n"
+"DOS partition table format can not be used on drives for volumes\n"
+"larger than (%llu bytes) for %d-byte sectors. Use parted(1) and GUID \n"
+"partition table format (GPT).\n\n"),
+			hectogiga / 10, hectogiga % 10,
+			total_number_of_sectors << 9,
+			(unsigned long long ) UINT_MAX * sector_size,
+			sector_size);
+	}
 }
 
 static void
@@ -891,11 +909,10 @@ get_partition_table_geometry(void) {
 
 void
 get_geometry(int fd, struct geom *g) {
-	int sec_fac;
-	unsigned long long llsectors, llcyls;
+	unsigned long long llcyls;
 
 	get_sectorsize(fd);
-	sec_fac = sector_size / 512;
+	sector_factor = sector_size / 512;
 	guess_device_type(fd);
 	heads = cylinders = sectors = 0;
 	kern_heads = kern_sectors = 0;
@@ -911,16 +928,14 @@ get_geometry(int fd, struct geom *g) {
 		pt_sectors ? pt_sectors :
 		kern_sectors ? kern_sectors : 63;
 
-	if (blkdev_get_sectors(fd, &llsectors) == -1)
-		llsectors = 0;
-
-	total_number_of_sectors = llsectors;
+	if (blkdev_get_sectors(fd, &total_number_of_sectors) == -1)
+		total_number_of_sectors = 0;
 
 	sector_offset = 1;
 	if (dos_compatible_flag)
 		sector_offset = sectors;
 
-	llcyls = total_number_of_sectors / (heads * sectors * sec_fac);
+	llcyls = total_number_of_sectors / (heads * sectors * sector_factor);
 	cylinders = llcyls;
 	if (cylinders != llcyls)	/* truncated? */
 		cylinders = ~0;
@@ -1458,7 +1473,7 @@ change_sysid(void) {
 	/* If sgi_label then don't use get_existing_partition,
 	   let the user select a partition, since get_existing_partition()
 	   only works for Linux like partition tables. */
-	if (!sgi_label) { 
+	if (!sgi_label) {
 		i = get_existing_partition(0, partitions);
 	} else {
 		i = get_partition(0, partitions);
@@ -1617,14 +1632,16 @@ list_disk_geometry(void) {
 	if (megabytes < 10000)
 		printf(_("\nDisk %s: %ld MB, %lld bytes\n"),
 		       disk_device, megabytes, bytes);
-	else
+	else {
+		long hectomega = (megabytes + 50) / 100;
 		printf(_("\nDisk %s: %ld.%ld GB, %lld bytes\n"),
-		       disk_device, megabytes/1000, (megabytes/100)%10, bytes);
+		       disk_device, hectomega / 10, hectomega % 10, bytes);
+	}
 	printf(_("%d heads, %llu sectors/track, %d cylinders"),
 	       heads, sectors, cylinders);
 	if (units_per_sector == 1)
 		printf(_(", total %llu sectors"),
-		       total_number_of_sectors / (sector_size/512));
+		       total_number_of_sectors / sector_factor);
 	printf("\n");
 	printf(_("Units = %s of %d * %d = %d bytes\n"),
 	       str_units(PLURAL),
@@ -1922,7 +1939,8 @@ check(int n, unsigned int h, unsigned int s, unsigned int c,
 static void
 verify(void) {
 	int i, j;
-	unsigned long total = 1;
+	unsigned long long total = 1;
+	unsigned long long n_sectors = (total_number_of_sectors / sector_factor);
 	unsigned long long first[partitions], last[partitions];
 	struct partition *p;
 
@@ -1985,12 +2003,12 @@ verify(void) {
 		}
 	}
 
-	if (total > total_number_of_sectors)
-		printf(_("Total allocated sectors %ld greater than the maximum"
-			" %lld\n"), total, total_number_of_sectors);
-	else if (total < total_number_of_sectors)
-		printf(_("%lld unallocated sectors\n"),
-		       total_number_of_sectors - total);
+	if (total > n_sectors)
+		printf(_("Total allocated sectors %llu greater than the maximum"
+			" %llu\n"), total, n_sectors);
+	else if (total < n_sectors)
+		printf(_("%lld unallocated %d-byte sectors\n"),
+		       n_sectors - total, sector_size);
 }
 
 static void
@@ -2014,7 +2032,7 @@ add_partition(int n, int sys) {
 		if (display_in_cyl_units || !total_number_of_sectors)
 			llimit = heads * sectors * cylinders - 1;
 		else
-			llimit = total_number_of_sectors - 1;
+			llimit = (total_number_of_sectors / sector_factor) - 1;
 		limit = llimit;
 		if (limit != llimit)
 			limit = 0x7fffffff;
@@ -2035,7 +2053,7 @@ add_partition(int n, int sys) {
 	do {
 		temp = start;
 		for (i = 0; i < partitions; i++) {
-			unsigned int lastplusoff;
+			unsigned long long lastplusoff;
 
 			if (start == ptes[i].offset)
 				start += sector_offset;
@@ -2051,7 +2069,7 @@ add_partition(int n, int sys) {
 			read = 0;
 		}
 		if (!read && start == temp) {
-			unsigned int i = start;
+			unsigned long long i = start;
 
 			start = read_int(cround(i), cround(i), cround(limit),
 					 0, mesg);
@@ -2263,18 +2281,21 @@ reread_partition_table(int leave) {
 
 	i = fstat(fd, &statbuf);
 	if (i == 0 && S_ISBLK(statbuf.st_mode)) {
-		printf(_("Calling ioctl() to re-read partition table.\n"));
 		sync();
 		sleep(2);
+#ifdef BLKRRPART
+		printf(_("Calling ioctl() to re-read partition table.\n"));
 		i = ioctl(fd, BLKRRPART);
+#else
+		errno = ENOSYS;
+		i = 1;
+#endif
         }
 
 	if (i) {
-		printf(_("\nWARNING: Re-reading the partition table "
-			 "failed with error %d: %s.\n"
-			 "The kernel still uses the old table.\n"
-			 "The new table will be used "
-			 "at the next reboot.\n"),
+		printf(_("\nWARNING: Re-reading the partition table failed with error %d: %s.\n"
+			 "The kernel still uses the old table. The new table will be used at\n"
+			 "the next reboot or after you run partprobe(8) or kpartx(8)\n"),
 			errno, strerror(errno));
 	}
 
@@ -2533,8 +2554,9 @@ try(char *device, int user_specified) {
 static void
 tryprocpt(void) {
 	FILE *procpt;
-	char line[100], ptname[100], devname[120];
-	int ma, mi, sz;
+	char line[128], ptname[128], devname[256];
+	int ma, mi;
+	unsigned long long sz;
 
 	procpt = fopen(PROC_PARTITIONS, "r");
 	if (procpt == NULL) {
@@ -2543,11 +2565,11 @@ tryprocpt(void) {
 	}
 
 	while (fgets(line, sizeof(line), procpt)) {
-		if (sscanf (line, " %d %d %d %[^\n ]",
+		if (sscanf (line, " %d %d %llu %128[^\n ]",
 			    &ma, &mi, &sz, ptname) != 4)
 			continue;
 		snprintf(devname, sizeof(devname), "/dev/%s", ptname);
-		if (is_probably_full_disk(devname))
+		if (is_whole_disk(devname))
 			try(devname, 0);
 	}
 	fclose(procpt);
@@ -2580,7 +2602,7 @@ main(int argc, char **argv) {
 	 *  fdisk [-b sectorsize] [-u] device
 	 *
 	 * Options -C, -H, -S set the geometry.
-	 * 
+	 *
 	 */
 	while ((c = getopt(argc, argv, "b:C:H:lsS:uvV")) != -1) {
 		switch (c) {
@@ -2591,7 +2613,7 @@ main(int argc, char **argv) {
 			*/
 			sector_size = atoi(optarg);
 			if (sector_size != 512 && sector_size != 1024 &&
-			    sector_size != 2048)
+			    sector_size != 2048 && sector_size != 4096)
 				fatal(usage);
 			sector_offset = 2;
 			user_set_sector_size = 1;
@@ -2601,7 +2623,7 @@ main(int argc, char **argv) {
 			break;
 		case 'H':
 			user_heads = atoi(optarg);
-			if (user_heads <= 0 || user_heads >= 256)
+			if (user_heads <= 0 || user_heads > 256)
 				user_heads = 0;
 			break;
 		case 'S':
